/*
 *  log10_linear_interpolate.c
 *  
 *
 *  Created by Rob Porritt on 2/22/11.
 *  Copyright 2011 UC Berkeley. All rights reserved.
 *
 *
 *  Simple code to interpolate a power spectrum
 *  Basically reads a psd file from compute_psd
 *  and then takes the output period array and
 *  converts it into some regular array which is
 *  user specified 
 */


//subroutine for interpolate.
//freqs is an array of frequencies for the input array tmp with nfreqs
//outputs to array out with nfreq number of frequencies
#include "station_analysis_tools.h"


int main (int argc, char *argv[]) {
	/* local variables */
	int x, l, xNode, xNodeOut;
	double minPeriodOut, maxPeriodOut;
	double *frequencyArray, *powerSpectralDensityRead, *powerSpectralDensityOut;
	double *standardDeviationRead, *standardDeviationOut, *splineFreqs;
	char paramFileName[100], inputFileName[100], outputFileName[100];
	char buff[100];
	FILE *paramFile, *inputFile, *outputFile;
	
	/* check usage */
	if ( argc != 2) {
		fprintf(stderr,"Usage: %s parameter_file\n",argv[0]);
		fprintf(stderr,"Parameter file format:\n");
		fprintf(stderr,"input_psd_file_name\n");
		fprintf(stderr,"number_of_nodes_in_x\n");
		fprintf(stderr,"min_period_out max_period_out period_increment\n");
		fprintf(stderr,"output_psd_file_name\n");
		exit(1);
	}
	
	/* -h help */
	if ( !(strcmp(argv[1],"-h"))) {
		fprintf(stderr,"Usage: %s parameter_file\n",argv[0]);
		fprintf(stderr,"Parameter file format:\n");
		fprintf(stderr,"input_psd_file_name\n");
		fprintf(stderr,"number_of_nodes_in_x\n");
		fprintf(stderr,"min_period_out max_period_out period_increment\n");
		fprintf(stderr,"output_psd_file_name\n");
		exit(1);
	}
	
	/* try to open the input file */
	sscanf(argv[1],"%s",paramFileName);
	if ( ( paramFile = fopen(paramFileName,"r")) == NULL) {
		fprintf(stderr,"Error, file %s not found\n",paramFileName);
		exit(1);
	}
	
	/* read the parameter file */
	l=0;
	while (fgets(buff,100,paramFile)) {
		if (l == 0) {
			sscanf(buff,"%s",inputFileName);
			l++;
		} else if (l == 1) {
			sscanf(buff,"%d",&xNode);
			l++;
		} else if (l == 2) {
			sscanf(buff,"%lf %lf",&minPeriodOut, &maxPeriodOut);
			l++;
		} else if (l == 3) {
			sscanf(buff,"%s",outputFileName);
			l++;
		}
	}
	
	/* close it up */
	fclose(paramFile);
	
	/* compute the output nodes */
	xNodeOut = compute_log10_nodes(maxPeriodOut, minPeriodOut);
		
	/* allocate memory */
	if ((splineFreqs = malloc(sizeof(*splineFreqs) * xNodeOut)) == NULL) {
		fprintf(stderr,"Error allocating splineFreqs\n");
		exit(1);
	}

	if ( ( frequencyArray = malloc(sizeof(*frequencyArray) * xNode)) == NULL) {
		fprintf(stderr,"Error allocating frequency array\n");
		exit(1);
	}
	
	if ( ( powerSpectralDensityRead = malloc(sizeof(*powerSpectralDensityRead) * xNode)) == NULL) {
		fprintf(stderr,"Error allocating psd array\n");
		exit(1);
	}
	
	if ( ( powerSpectralDensityOut = malloc(sizeof(*powerSpectralDensityOut) * xNodeOut)) == NULL) {
		fprintf(stderr,"Error allocating psd output array\n");
		exit(1);
	}
	
	if ( ( standardDeviationRead = malloc(sizeof(*standardDeviationRead) * xNode)) == NULL) {
		fprintf(stderr,"Error allocating sd array\n");
		exit(1);
	}
	
	if ( ( standardDeviationOut = malloc(sizeof(*standardDeviationOut) * xNodeOut)) == NULL) {
		fprintf(stderr,"Error allocating sd output array\n");
		exit(1);
	}
	
	/* try to open the input file */
	if ( ( inputFile = fopen(inputFileName,"r")) == NULL) {
		fprintf(stderr,"Error, could not read inputFileName: %s\n",inputFileName);
		exit(1);
	}
	
	/* read the file */
	for (x=xNode-1; x>=0; x--) {
		fscanf(inputFile,"%lf %lf %lf",&frequencyArray[x],&powerSpectralDensityRead[x],&standardDeviationRead[x]);
	}
		
	/* close the file */
	fclose(inputFile);
	
	//do the interpolation
	log10_linear_interpolate(frequencyArray, xNode, powerSpectralDensityRead, splineFreqs, powerSpectralDensityOut, &xNodeOut, minPeriodOut, maxPeriodOut);
	log10_linear_interpolate(frequencyArray, xNode, standardDeviationRead, splineFreqs, standardDeviationOut, &xNodeOut, minPeriodOut, maxPeriodOut);
		
	/* open the output file */
	if ( ( outputFile = fopen(outputFileName,"w")) == NULL) {
		fprintf(stderr,"Error, could not create file %s\n",outputFileName);
		exit(1);
	}
	
	/* print */
	for (x=xNodeOut-1; x>=0; x--) {
		/* avoid extrapoloation */
		if (splineFreqs[x] >= frequencyArray[0] && splineFreqs[x] <= frequencyArray[xNode-1]) {
			fprintf(outputFile,"%3.8g %3.8g %3.8g\n",splineFreqs[x],powerSpectralDensityOut[x],standardDeviationOut[x]);
		}
	}
	
	/* close up */
	fclose(outputFile);
	
	/* because we write to a hard wired file the spline part, let the user know */
	printf("log10 based linear interpolated file: %s\n",outputFileName);
	
	/* why not free up memory */
	free(frequencyArray);
	free(splineFreqs);
	free(powerSpectralDensityOut);
	free(standardDeviationOut);
	free(powerSpectralDensityRead);
	free(standardDeviationRead);
	
	return 0;
}





