/*
 *  mspline.c
 *  
 *
 *  Created by Rob Porritt on 12/27/10.
 *  Copyright 2010 UC Berkeley. All rights reserved.
 *
 *	C code to replicate the mspline.f codes. Once finished, they should be able to be copied directly into lib_ftan.c
 *  a few define statements will be necessary and they are already in the ftan.h header
 */
#include "spline.h"

/* simple function they define for each subroutine. Syntax used in subroutines looks more like array syntax, so here defined as a simple function */
double fsig(double a, double b, double c) {
	return (a-c) * (b-c);
}

/* more or less the main driver spline function.
 * checks boundary conditions, calls cubspl, and cleans things up a bit
 * return value is an error code
 * 0 indicates success
 */
int mspline(int ip, int n, double *x, double *y, int ind1, double d1, int ind2, double d2, SPLINECOMMON *splineCommon) {
	/* subroutine variables. */
	int i, j;
	double s,z;
	double **c, **cc;
	
	/* error checking */
	if (n<2 || n >= MAXSPLINE) {
		fprintf(stderr,"Error in mspline. n=%d. Acceptable values are between 2 and %1.0lf. Returning Error code: 1\n",n,MAXSPLINE);
		return 1;
	}
	
	if (ip < 0 || ip >= MINSPLINE) {
		fprintf(stderr,"Error in mspline. ip=%d. Acceptable values are between 0 and %1.0lf. Returning Error code: 2\n",ip, MINSPLINE);
		return 2;
	}
	
	if (ind1 < 0 || ind1 > 2) {
		fprintf(stderr,"Error in mspline. ind1=%d. Acceptable values are between 0 and 2. Returning Error code: 3\n",ind1);
		return 3;
	}
	
	if (ind2 < 0 || ind2 > 2) {
		fprintf(stderr,"Error in mspline. ind2=%d. Acceptable values are between 0 and 2. Returning Error code: 4\n",ind2);
		return 4;
	}
	
	/* allocate a little memory for c - don't forget to free it */
	c = malloc(sizeof(*c) * 4);
	for (i=0; i<4; i++) {
		c[i] = malloc(sizeof(*c) * n);
	}
	
	cc = malloc(sizeof(*cc) * 5);
	for (i=0; i<5; i++) {
		cc[i] = malloc(sizeof(*cc) * n);
	}
	
	
	/* fill the first part of the c array */
	for (i=0; i<n; i++) {
		c[0][i] = y[i];
	}
	
	/* fill the end points of the second c array */
	c[1][0] = d1;
	c[1][n-1] = d2;
	/* call to cubic spline */
	cubspl(x, c, n, ind1, ind2, splineCommon);
	
	/* remove 2 and 6? */
	for (i=0; i<n-1; i++) {
		c[2][i] = c[2][i] / 2.0;
		c[3][i] = c[3][i] / 6.0;
	}
	
	/* create integral coefficients */
	s = 0.0;
	for (i=0; i<n-1; i++) {
		for (j=0; j<4; j++) {
			cc[j+1][i] = c[j][i] / j;
		}
		cc[0][i] = s;
		z = x[i+1] - x[i];
		s = s + z * (cc[1][i] + z * (cc[2][i] + z * (cc[3][i] + z * cc[4][i])));
	}
	
	/* save the coefficients into common block variables */
	splineCommon->nn[ip] = n;
	for (i=0; i<n; i++) {
		splineCommon->xx[i][ip] = x[i];
		for (j=0; j<4; j++) {
			splineCommon->sc[j][i][ip] = c[j][i];
			splineCommon->scc[j][i][ip] = cc[j][i];
		}
		splineCommon->scc[4][i][ip] = cc[4][i];
	}
	
	/* freeing the c array allocated earlier */
	for (i=0; i<4; i++) {
		free(c[i]);
	}
	free(c);
	
	for (i=0; i<5; i++) {
		free(cc[i]);
	}
	free(cc);
	
	
	/* done */
	return 0;
}

/* 
 * integral for spline
 * ...don't really ever see this called, but maybe its needed.
 * after finishing the coding, there is no use of this in the ftan code.
 * Potentially useful for adaptation in future spline endeavors
 */
int msplint(int ip, double sa, double sb, double *sint, SPLINECOMMON *splineCommon) {
	int ierr, i, ii, j, n;
	double x1, x2, r, s1, s2, z;
	double *x, **cc;
	
	/* inially there is no error (woot!) */
	ierr = 0;
	
	x = malloc(sizeof(*x) * splineCommon->nn[ip]);
	cc = malloc(sizeof(*cc) * 5);
	for (i=0; i<5; i++) {
		cc[i] = malloc(sizeof(*cc) * splineCommon->nn[ip]);
	}
	
	/* restore polynomial coefficients */
	n = splineCommon->nn[ip];
	for (i=0; i<n; i++) {
		x[i] = splineCommon->xx[i][ip];
		for (j=0; j<5; j++) {
			cc[j][i] = splineCommon->scc[j][i][ip];
		}
	}
	
	/* compute integral for sa */
	ii = 0;
	for (i=0; i<n-1; i++) {
		if (fsig(x[i],x[i+1],sa) <= 0.0) {
			ii = i;
		}
	}
	if (ii == 0) {
		if (sa <= x[0]) {
			ii = 0;
		}
		if (sa >= x[n-2]) {
			ii = n-2;
		}
	}
	z = sa-x[ii];
	s1 = cc[0][ii] + z * (cc[1][ii] + z * (cc[2][ii] + z * (cc[3][ii] + z * cc[4][ii])));
	
	/* compute integral for sb */
	ii = 0;
	for (i=0; i<n-1; i++) {
		if (fsig(x[i],x[i+1],sb) <= 0.0) {
			ii=i;
		}
	}
	if (ii == 0) {
		if (sb <= x[0]) {
			ii = 0;
		}
		if (sb >= x[n-2]) {
			ii = n-2;
		}
	}
	z = sb - x[ii];
	s2 = cc[0][ii] + z * (cc[1][ii] + z * (cc[2][ii] + z * (cc[3][ii] + z * cc[4][ii])));
	
	/* appears like we probably want the value computed as sint. Thus we make it a pointer. */
	*sint = s2 - s1;

	free(x);
	for (i=0; i<5; i++) {
		free(cc[i]);
	}
	free(cc);
	
	return ierr;
}

/*
 * spline interpolation function and its 1st and 2nd derivatives
 */
int msplder(int ip, double xt, double *s, double *sd, double *sdd, SPLINECOMMON *splineCommon) {
	int ierr, i, j, ii, n;
	double *x, **c;
	double x1, x2, r, z;
	
	/* initially no error (woot!) */
	ierr = 0;
	
	x = malloc(sizeof(*x) * splineCommon->nn[ip]);
	c = malloc(sizeof(*c) * 4);
	for (i=0; i<4; i++) {
		c[i] = malloc(sizeof(*c) * splineCommon->nn[ip]);
	}
	
	
	/* restore polynomial coefficients */
	n = splineCommon->nn[ip];
	for (i=0; i<n; i++) {
		x[i] = splineCommon->xx[i][ip];
		for (j=0; j<4; j++) {
			c[j][i] = splineCommon->sc[j][i][ip];
		}
	}
	
	/* find interval for interpolation */
	ii = 0;
	for (i=0; i<n-1; i++) {
		if (fsig(x[i],x[i+1],xt) <= 0.0) {
			ii = i;
		}
	}
	if (ii == 0) {
		if (xt <= x[0]) {
			ii = 0;
		}
		if (xt >= x[n-2]) {
			ii = n-2;
		}
	}
	z = xt - x[ii];
	
	/* the computed values are pointers so we can return the altered values */
	*s = c[0][ii]+z*(c[1][ii]+z*(c[2][ii]+z*c[3][ii]));
	*sd = c[1][ii]+z*(2.0*c[2][ii]+3.0*z*c[3][ii]);
	*sdd = 2.0*c[2][ii]+6.0*z*c[3][ii];
	
	free(x);
	for (i=0; i<4; i++) {
		free(c[i]);
	}
	free(c);
	
	
	return ierr;
}

/* 
 * the big nasty. cubspl() uses a maze of goto statements. 
 */

void cubspl(double *tau, double **c, int n, int ibcbeg, int ibcend, SPLINECOMMON *splineCommon) {
	/*
	 Original comments:
	 c  from  * a practical guide to splines *  by c. de boor    
	 c     ************************  input  ***************************
	 c     n = number of data points. assumed to be greater than 1.
	 c     (tau(i), c(1,i), i=1,...,n) = abscissae and ordinates of the
	 c        data points. tau is assumed to be strictly increasing.
	 c     ibcbeg, ibcend = boundary condition indicators, and
	 c     c(2,1), c(2,n) = boundary condition information. Specifically,
	 c        ibcbeg = 0  means no boundary condition at tau(1) is given.
	 c           in this case, the not-a-knot condition is used, i.e. the
	 c           jump in the third derivative across tau(2) is forced to
	 c           zero, thus the first and the second cubic polynomial pieces
	 c           are made to coincide.)
	 c        ibcbeg = 1  means that the slope at tau(1) is made to equal
	 c           c(2,1), supplied by input.
	 c        ibcbeg = 2  means that the second derivative at tau(1) is
	 c           made to equal c(2,1), supplied by input.
	 c        ibcend = 0, 1, or 2 has analogous meaning concerning the
	 c           boundary condition at tau(n), with the additional infor-
	 c           mation taken from c(2,n).
	 c     ***********************  output  **************************
	 c     c(j,i), j=1,...,4; i=1,...,l (= n-1) = the polynomial coefficients
	 c        of the cubic interpolating spline with interior knots (or
	 c        joints) tau(2), ..., tau(n-1). precisely, in the interval
	 c        (tau(i), tau(i+1)), the spline f is given by
	 c           f(x) = c(1,i)+h*(c(2,i)+h*(c(3,i)+h*c(4,i)/3.)/2.)
	 c        where h = x - tau(i). the function program *ppvalu* may be
	 c        used to evaluate f or its derivatives from tau,c, l = n-1,
	 c        and k=4.
	 implicit none
	 c ---
	 integer*4 ibcbeg,ibcend,n,i,j,l,m
	 real*8    c(4,n),tau(n),divdf1,divdf3,dtau,g
	 c ---
	 c****** a tridiagonal linear system for the unknown slopes s(i) of
	 c  f  at tau(i), i=1,...,n, is generated and then solved by gauss elim-
	 c  ination, with s(i) ending up in c(2,i), all i.
	 c     c(3,.) and c(4,.) are used initially for temporary storage.
	 l = n-1
	 c compute first differences of tau sequence and store in c(3,.). also,
	 c compute first divided difference of data and store in c(4,.).
	 do m=2,n
	 c(3,m) = tau(m) - tau(m-1)
	 c(4,m) = (c(1,m) - c(1,m-1))/c(3,m)
	 enddo
	 c construct first equation from the boundary condition, of the form
	 c             c(4,1)*s(1) + c(3,1)*s(2) = c(2,1)
	 c      if (ibcbeg-1) 11,15,16
	 if (ibcbeg - 1 .LT. 0) GOTO 11
	 if (ibcbeg - 1 .EQ. 0) GOTO 15
	 if (ibcbeg - 1 .GT. 0) GOTO 16
	 11 if (n .gt. 2) goto 12
	 c     no condition at left end and n = 2.
	 c(4,1) = 1.0d0
	 c(3,1) = 1.0d0
	 c(2,1) = 2.0d0*c(4,2)
	 goto 25
	 c     not-a-knot condition at left end and n .gt. 2.
	 12 c(4,1) = c(3,3)
	 c(3,1) = c(3,2) + c(3,3)
	 c(2,1) =((c(3,2)+2.0d0*c(3,1))*c(4,2)*c(3,3)+c(3,2)**2*c(4,3))/c(3,1)
	 goto 19
	 c     slope prescribed at left end.
	 15 c(4,1) = 1.0d0
	 c(3,1) = 0.0d0
	 goto 18
	 c     second derivative prescribed at left end.
	 16 c(4,1) = 2.0d0
	 c(3,1) = 1.0d0
	 c(2,1) = 3.0d0*c(4,2) - c(3,2)/2.0d0*c(2,1)
	 18 if(n .eq. 2) goto 25
	 c  if there are interior knots, generate the corresp. equations and car-
	 c  ry out the forward pass of gauss elimination, after which the m-th
	 c  equation reads    c(4,m)*s(m) + c(3,m)*s(m+1) = c(2,m).
	 19 do m=2,l
	 g = -c(3,m+1)/c(4,m-1)
	 c(2,m) = g*c(2,m-1) + 3.0d0*(c(3,m)*c(4,m+1)+c(3,m+1)*c(4,m))
	 c(4,m) = g*c(3,m-1) + 2.0d0*(c(3,m) + c(3,m+1))
	 enddo
	 c construct last equation from the second boundary condition, of the form
	 c           (-g*c(4,n-1))*s(n-1) + c(4,n)*s(n) = c(2,n)
	 c     if slope is prescribed at right end, one can go directly to back-
	 c     substitution, since c array happens to be set up just right for it
	 c     at this point.
	 c      if (ibcend-1) 21,30,24
	 if (ibcend - 1 .LT. 0) GOTO 21
	 if (ibcend - 1 .EQ. 0) GOTO 30
	 if (ibcend - 1 .GT. 0) GOTO 24
	 21 if (n .eq. 3 .and. ibcbeg .eq. 0) goto 22
	 c     not-a-knot and n .ge. 3, and either n.gt.3 or  also not-a-knot at
	 c     left end point.
	 g = c(3,n-1) + c(3,n)
	 c(2,n) = ((c(3,n)+2.0d0*g)*c(4,n)*c(3,n-1)
     *            + c(3,n)**2*(c(1,n-1)-c(1,n-2))/c(3,n-1))/g
	 g = -g/c(4,n-1)
	 c(4,n) = c(3,n-1)
	 goto 29
	 c     either (n=3 and not-a-knot also at left) or (n=2 and not not-a-
	 c     knot at left end point).
	 22 c(2,n) = 2.0d0*c(4,n)
	 c(4,n) = 1.0d0
	 goto 28
	 c     second derivative prescribed at right endpoint.
	 24 c(2,n) = 3.0d0*c(4,n) + c(3,n)/2.0d0*c(2,n)
	 c(4,n) = 2.0d0
	 goto 28
	 c   25 if (ibcend-1) 26,30,24
	 25 if (ibcend-1 .LT. 0) then
	 goto 26
	 else if (ibcend-1 .EQ. 0) then
	 goto 30
	 else if (ibcend-1 .GT. 0) then
	 goto 24
	 end if
	 
	 26 if (ibcbeg .gt. 0) goto 22
	 c     not-a-knot at right endpoint and at left endpoint and n = 2.
	 c(2,n) = c(4,n)
	 goto 30
	 28 g = -1.0d0/c(4,n-1)
	 c complete forward pass of gauss elimination.
	 29 c(4,n) = g*c(3,n-1) + c(4,n)
	 c(2,n) = (g*c(2,n-1) + c(2,n))/c(4,n)
	 c ccarry out back substitution
	 30 j = l 
	 40 c(2,j) = (c(2,j) - c(3,j)*c(2,j+1))/c(4,j)
	 j = j - 1
	 if (j .gt. 0) goto 40
	 c****** generate cubic coefficients in each interval, i.e., the deriv.s
	 c  at its left endpoint, from value and slope at its endpoints.
	 do i=2,n
	 dtau = c(3,i)
	 divdf1 = (c(1,i) - c(1,i-1))/dtau
	 divdf3 = c(2,i-1) + c(2,i) - 2.0d0*divdf1
	 c(3,i-1) = 2.0d0*(divdf1 - c(2,i-1) - divdf3)/dtau
	 c(4,i-1) = (divdf3/dtau)*(6.0d0/dtau)
	 enddo
	 return
	 end	 
	*/
	int i, j, l, m, niner_flag=0, twenty_nine_flag = 0, twenty_eight_flag = 0;
	double divdf1, divdf3, dtau, g;
	
	/*
	 c****** a tridiagonal linear system for the unknown slopes s[i] of
	 c  f  at tau[i], i=0,...,n-1, is generated and then solved by gauss elim-
	 c  ination, with s[i] ending up in c[1][i], all i.
	 c     c[2][.] and c[3][.] are used initially for temporary storage.
	*/
	l = n-1;
	
	/* compute first differences of tau sequence and store in c[2][.]. Also,
	 * compute the first divided difference of data and store in c[3][.];
	 * note that c and tau were entered as properly allocated vectors;
	 */
	for (m=1; m<n; m++) {
		c[2][m] = tau[m] - tau[m-1];
		c[3][m] = (c[0][m] - c[0][m-1]) / c[2][m];
	}
	
	/* 
	 * construct first equation from the boundary condition. It has the form:
	 *			c[3][0] * s[0] + c[2][0] * s[1] = c[1][0]
	 */
	/* construct the left end conditions. */
	if (ibcbeg - 1 < 0) {
		if (n <= 2) { // no condition at left end and n = 2
			c[3][0] = 1.0;
			c[2][0] = 1.0;
			c[1][0] = 2.0 * c[3][1];
			niner_flag = 0;
			//goto 25...
		} else { //not-a-knot condition at left end and n > 2
			c[3][0] = c[2][2];
			c[2][0] = c[2][1] + c[2][2];
			c[1][0] = ((c[2][1] + 2.0 * c[2][0]) * c[3][1] * c[2][2] + c[2][1] * c[2][1] * c[3][2]) / c[2][0];
			// goto 19 - check for interior knots...
			niner_flag = 1;
		}
	} else if ( ibcbeg - 1 == 0) {
		c[3][0] = 1.0;
		c[2][0] = 0.0;
		if (n == 2) {
			niner_flag = 0;
		} else {
			niner_flag = 1;
		}
	} else {
		c[3][0] = 2.0;
		c[2][0] = 1.0;
		c[1][0] = 3.0 * c[3][1] - c[2][1] / 2.0 * c[1][0];
		if (n == 2) {
			niner_flag = 0;
		} else {
			niner_flag = 1;
		}
	}
	
	/* do the steps in goto 19 */
	if (niner_flag == 1) {
		for (m=1; m<l; m++) {
			g = -1.0 * c[2][m+1] / c[3][m-1];
			c[1][m] = g*c[1][m-1] + 3.0 * (c[2][m] * c[3][m+1] + c[2][m+1] * c[3][m]);
			c[3][m] = g*c[2][m-1] + 2.0 * (c[2][m] + c[2][m+1]);
		}
		/* this would be skipped by the "goto 25" statement which has been changed to niner_flag = 0 */
		 if (ibcend - 1 < 0) {
			 if (n != 3 || ibcbeg != 0) {
				 g = c[2][n-2] + c[2][n-1];
				 c[1][n-1] = ((c[2][n-1] + 2.0 * g) * c[3][n-1] * c[2][n-2] + c[2][n-1] * c[2][n-1] * (c[0][n-2]-c[0][n-3]) / c[2][n-2]) / g;
				 g = -1.0 * g / c[3][n-2];
				 c[3][n-1] = c[2][n-2];
				 twenty_eight_flag = 0;
				 twenty_nine_flag = 1;
			 } else {
				 c[1][n-1] = 2.0 * c[3][n-1];
				 c[3][n-1] = 1.0;
				 twenty_eight_flag = 1;
			 } 
		 } else if (ibcend - 1 == 0) {
			 twenty_nine_flag = 0;
			 twenty_eight_flag = 0;
		 } else if (ibcend - 1 > 0) {
			 c[1][n-1] = 3.0 * c[3][n-1] + c[2][n-1] / 2.0 * c[1][n-1];
			 c[3][n-1] = 2.0;
			 twenty_eight_flag = 1;
		 }
	} else {
		if (ibcend - 1 < 0) {
			if (ibcbeg < 0) {
				c[1][n-1] = c[3][n-1];
				twenty_eight_flag = 0;
				twenty_nine_flag = 0;
			} else {
				c[1][n-1] = 2.0 * c[3][n-1];
				c[3][n-1] = 1.0;
				twenty_eight_flag = 1;
			}
		} else if (ibcend -1 > 0) {
			c[1][n-1] = 3.0 * c[3][n-1] + c[2][n-1] / 2.0 * c[1][n-1];
			c[3][n-1] = 2.0;
			twenty_eight_flag = 1;
		} else {
			twenty_eight_flag = 0;
			twenty_nine_flag = 0;
		}
	}
	
	//28
	if (twenty_eight_flag == 1) {
		g = -1.0 / c[3][n-2];
		twenty_nine_flag = 1;
	}
	//29
	if (twenty_nine_flag == 1) {
		c[3][n-1] = g * c[2][n-2] + c[3][n-1];
		c[1][n-1] = (g*c[1][n-2] + c[1][n-1]) / c[3][n-1];
	}
	
	/* carry out back substitution */
	for (j=l-1; j>=0; j--) {
		c[1][j] = (c[1][j] - c[2][j] * c[1][j+1]) / c[3][j];
	}
	
	/* generate cubic coefficients in each interval.
	 * ie the derivatives at its left endpoint 
	 * from value and slope at its endpoints
	 */
	for (i=1; i<n; i++) {
		dtau = c[2][i];
		divdf1 = (c[0][i] - c[0][i-1]) / dtau;
		divdf3 = c[1][i-1] + c[1][i] - 2.0 * divdf1;
		c[2][i-1] = 2.0 * ( divdf1 - c[1][i-1] - divdf3) / dtau;
		c[3][i-1] = (divdf3 / dtau) * (6.0 / dtau);
	}
	
	return;
}


void allocate_spline(SPLINECOMMON *splineCommon, int nmin, int nmax) {
	/*
	 * typedef struct {
	 * int nn[MINSPLINE];
	 * int ndim[MINSPLINE];
	 * double xx[MAXSPLINE][MINSPLINE];
	 * double sc[4][MAXSPLINE][MINSPLINE];
	 * double scc[5][MAXSPLINE][MINSPLINE];
	 * } SPLINECOMMON;
	 */
	int i, j;
	
	//ints
	splineCommon->nn = malloc(sizeof(splineCommon->nn) * nmin);
	splineCommon->ndim = malloc(sizeof(splineCommon->ndim) * nmax);
	
	//doubles
	splineCommon->xx = malloc(sizeof(splineCommon->xx) * nmax);
	for (i=0; i<nmax; i++) {
		splineCommon->xx[i] = malloc(sizeof(splineCommon->xx) * nmin);
	}
	
	splineCommon->sc = malloc(sizeof(splineCommon->sc) * 4);
	for (i=0; i<4; i++) {
		splineCommon->sc[i] = malloc(sizeof(splineCommon->sc) * nmax);
		for (j=0; j<nmax; j++) {
			splineCommon->sc[i][j] = malloc(sizeof(splineCommon->sc) * nmin);
		}
	}
	
	splineCommon->scc = malloc(sizeof(splineCommon->scc) * 5);
	for (i=0; i<5; i++) {
		splineCommon->scc[i] = malloc(sizeof(splineCommon->scc) * nmax);
		for (j=0; j<nmax; j++) {
			splineCommon->scc[i][j] = malloc(sizeof(splineCommon->scc) * nmin);
		}
	}
	
	return;
}

void free_spline(SPLINECOMMON *splineCommon, int nmin, int nmax) {
	int i,j;
	
	//ints
	free(splineCommon->nn);
	free(splineCommon->ndim);
	
	//doubles
	for (i=0; i<nmax; i++) {
		free(splineCommon->xx[i]);
	}
	free(splineCommon->xx);
	
	for (i=0; i<4; i++) {
		for (j=0; j<nmax; j++) {
			free(splineCommon->sc[i][j]);
		}
		free(splineCommon->sc[i]);
	}
	free(splineCommon->sc);
	
	for (i=0; i<5; i++) {
		for (j=0; j<nmax; j++) {
			free(splineCommon->scc[i][j]);
		}
		free(splineCommon->scc[i]);
	}
	free(splineCommon->scc);
	
	return;
}




