/*
general_pdf.c
uses user specified information to compute probability density function of whatever xy data is given.
must specify:
list of xy files - each with constant x-domain and same order of y-domain
n_x - number of x components. The max and min are found in the input files and the output array uses what's given as input
ymin ymax yinc - the y domain is relatively similar between all inputs, but this determines the output search space
output_file - an output file to return the data. It is gmt ready with data in x y z format.

This should be useful for a number of problems. Original design is for use with power spectral density, coherence, and incoherence while also working for their associated standard deviations. It may be interesting to also use with ambient noise cross-correlation measurements


*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

int main (int argc, char *argv[]) {
	/* local variables */
	int i, j, l, n_x, y_node;
	unsigned int **hit_matrix, n_files = 0;
	double ymin, ymax, yinc;	
	double *x_array, value, dummy;
	double y_top, y_bot;
	char paramfile_name[100], list_file_name[100], temp_file_name[100];
	char output_file_name[100], buff[100], buff2[100];
	FILE *paramfile, *output_file, *temp_file, *list_file;

	/* check usage */
	if (argc != 2) {
		fprintf(stderr,"Usage: %s parameter_file\n", argv[0]);
		fprintf(stderr,"parameter_file format: \n");
		fprintf(stderr,"list_of_input_files\n");
		fprintf(stderr,"n_x\n");
		fprintf(stderr,"ymin ymax yinc\n");
		fprintf(stderr,"output_file\n");
		exit(1);
	}

	/* -h help */
        if (!(strcmp(argv[1],"-h"))) {
                fprintf(stderr,"Usage: %s parameter_file\n", argv[0]);
                fprintf(stderr,"parameter_file format: \n");
                fprintf(stderr,"list_of_input_files\n");
                fprintf(stderr,"n_x\n");
                fprintf(stderr,"ymin ymax yinc\n");
                fprintf(stderr,"output_file\n");
                exit(1);
        }

	/* open parameter file */
	sscanf(argv[1],"%s",paramfile_name);
	if (! (paramfile = fopen(paramfile_name,"r"))) {
		fprintf(stderr,"Error, file %s not found\n",paramfile_name);
		exit(1);
	}

	/* read parameter file */
	l=0;
	while (fgets(buff,100,paramfile)) {
		if (l == 0) {
			sscanf(buff,"%s",list_file_name);
			l++;
		} else if (l == 1) {
			sscanf(buff,"%d",&n_x);
			l++;
		} else if (l == 2) {
			sscanf(buff,"%lf %lf %lf", &ymin, &ymax, &yinc);
			l++;
		} else if (l == 3) {
			sscanf(buff,"%s",output_file_name);
			l++;
		}
	}

	/* close parameter file */
	fclose(paramfile);

	/* compute y bin size */
	y_node = floor((ymax - ymin) / yinc + 1.5);

	/* allocate memory */
	x_array = malloc(sizeof(*x_array) * n_x);
	hit_matrix = malloc(sizeof(*hit_matrix) * n_x);
	for (i=0; i<n_x; i++) {
		hit_matrix[i] = malloc(sizeof(*hit_matrix) * y_node);
	}

	/* be sure the hit matrix is initialized to 0 */
	for (i=0; i<n_x; i++) {
		for (j=0; j<y_node; j++) {
			hit_matrix[i][j] = 0;
		}
	}

	/* try to open the list file */
	if (! (list_file = fopen(list_file_name,"r"))) {
		fprintf(stderr,"Error file %s not found\n",list_file_name);
		exit(1);
	}

	/* loop over each file in the list */
	while (fgets(buff2,100,list_file)) {
		
		/* scan the buffer to get the current file name */
		sscanf(buff2,"%s",temp_file_name);

		/* open the data file */
		if (! (temp_file = fopen(temp_file_name,"r"))) {
			printf("file %s not found\n", temp_file_name);
		} else {
			/* a file has been found, so read x y */
			n_files++;
			i = 0;
			while (fgets(buff,100,temp_file)) {
				/* only get the x array from the first file */
				if (n_files == 1) {
					sscanf(buff,"%lf %lf", &x_array[i], &value);
				} else {
					sscanf(buff,"%lf %lf", &dummy, &value);
				}
				
				/* search through the y-space and increase the hit_matrix where this value falls */
				for (j=0; j<y_node; j++) {
					/* define the top and bottom values for the search */ 
					y_bot = j * yinc + ymin;
					y_top = y_bot + yinc;
					if (value <= y_top && value >= y_bot) {
						hit_matrix[i][j]++;
						break;
					}
				}
				i++;
			}
			fclose(temp_file);
		}
	}

	/* close the list file */
	fclose(list_file);

	/* open the output file */
	output_file = fopen(output_file_name,"w");

	/* write the output */
	for (i=0; i<n_x; i++) {
		for (j=0; j<y_node; j++) {
			fprintf(output_file,"%lf %lf %lf\n",x_array[i], (double) j*yinc+ymin, (double) hit_matrix[i][j] / (double) n_files);
		}
	}

	/* close the output */
	fclose(output_file);

	/* free up memory */
	free(x_array);
	for (i=0; i<n_x; i++) {
		free(hit_matrix[i]);
	}
	free(hit_matrix);

	/* done message */
	fprintf(stderr,"Done with file: %s\n",output_file_name);

	return 0;
}




