/*
 *  interpolate_psd.c
 *  
 *
 *  Created by Rob Porritt on 10/1/10.
 *  Copyright 2010 UC Berkeley. All rights reserved.
 *
 *  Simple code to interpolate a power spectrum
 *  Basically reads a psd file from compute_psd
 *  and then takes the output period array and
 *  converts it into some regular array which is
 *  user specified
 *
 */


//subroutine for interpolate.
//freqs is an array of frequencies for the input array tmp with nfreqs
//outputs to array out with nfreq number of frequencies
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif


int main (int argc, char *argv[]) {
	/* local variables */
	int x, l, xNode, xNodeOut;
	float minPeriodOut, maxPeriodOut, periodIncrement;
	float *frequencyArray, *powerSpectralDensityRead, *powerSpectralDensityOut;
	float *standardDeviationRead, *standardDeviationOut;
	char paramFileName[100], inputFileName[100], outputFileName[100];
	char buff[100];
	FILE *paramFile, *inputFile, *outputFile;
	
	/* check usage */
	if ( argc != 2) {
		fprintf(stderr,"Usage: %s parameter_file\n",argv[0]);
		fprintf(stderr,"Parameter file format:\n");
		fprintf(stderr,"input_psd_file_name\n");
		fprintf(stderr,"number_of_nodes_in_x\n");
		fprintf(stderr,"min_period_out max_period_out period_increment\n");
		fprintf(stderr,"output_psd_file_name\n");
		exit(1);
	}
	
	/* -h help */
	if ( !(strcmp(argv[1],"-h"))) {
		fprintf(stderr,"Usage: %s parameter_file\n",argv[0]);
		fprintf(stderr,"Parameter file format:\n");
		fprintf(stderr,"input_psd_file_name\n");
		fprintf(stderr,"number_of_nodes_in_x\n");
		fprintf(stderr,"min_period_out max_period_out period_increment\n");
		fprintf(stderr,"output_psd_file_name\n");
		exit(1);
	}
	
	/* try to open the input file */
	sscanf(argv[1],"%s",paramFileName);
	if ( ( paramFile = fopen(paramFileName,"r")) == NULL) {
		fprintf(stderr,"Error, file %s not found\n",paramFileName);
		exit(1);
	}
	
	/* read the parameter file */
	l=0;
	while (fgets(buff,100,paramFile)) {
		if (l == 0) {
			sscanf(buff,"%s",inputFileName);
			l++;
		} else if (l == 1) {
			sscanf(buff,"%d",&xNode);
			l++;
		} else if (l == 2) {
			sscanf(buff,"%f %f %f",&minPeriodOut, &maxPeriodOut, &periodIncrement);
			l++;
		} else if (l == 3) {
			sscanf(buff,"%s",outputFileName);
			l++;
		}
	}

	/* close it up */
	fclose(paramFile);
	
	/* compute the output nodes */
//	xNodeOut = compute_nodes(maxPeriodOut, minPeriodOut, periodIncrement);
	xNodeOut = floor((maxPeriodOut - minPeriodOut) / periodIncrement + 1.5);
	
	/* allocate memory */
	if ( ( frequencyArray = malloc(sizeof(*frequencyArray) * xNode)) == NULL) {
		fprintf(stderr,"Error allocating frequency array\n");
		exit(1);
	}
	
	if ( ( powerSpectralDensityRead = malloc(sizeof(*powerSpectralDensityRead) * xNode)) == NULL) {
		fprintf(stderr,"Error allocating psd array\n");
		exit(1);
	}
	
	if ( ( powerSpectralDensityOut = malloc(sizeof(*powerSpectralDensityOut) * xNodeOut)) == NULL) {
		fprintf(stderr,"Error allocating psd output array\n");
		exit(1);
	}
	
	if ( ( standardDeviationRead = malloc(sizeof(*standardDeviationRead) * xNode)) == NULL) {
		fprintf(stderr,"Error allocating sd array\n");
		exit(1);
	}
	
	if ( ( standardDeviationOut = malloc(sizeof(*standardDeviationOut) * xNodeOut)) == NULL) {
		fprintf(stderr,"Error allocating sd output array\n");
		exit(1);
	}
	
	/* try to open the input file */
	if ( ( inputFile = fopen(inputFileName,"r")) == NULL) {
		fprintf(stderr,"Error, could not read inputFileName: %s\n",inputFileName);
		exit(1);
	}
		
	/* read the file */
	for (x=xNode-1; x>=0; x--) {
		fscanf(inputFile,"%f %f %f",&frequencyArray[x],&powerSpectralDensityRead[x],&standardDeviationRead[x]);
	}
	
	/* close the file */
	fclose(inputFile);
		
	/* call the interpolation routine */
	linear_interpolate_psd(frequencyArray, xNode, powerSpectralDensityRead,powerSpectralDensityOut, minPeriodOut, maxPeriodOut,periodIncrement);
	linear_interpolate_psd(frequencyArray, xNode, standardDeviationRead,standardDeviationOut, minPeriodOut, maxPeriodOut, periodIncrement);
	
	/* open the output file */
	if ( ( outputFile = fopen(outputFileName,"w")) == NULL) {
		fprintf(stderr,"Error, could not create file %s\n",outputFileName);
		exit(1);
	}
	
	/* print */
	for (x=xNodeOut-1; x>=0; x--) {
		fprintf(outputFile,"%3.8g %3.8g %3.8g\n",minPeriodOut+x*periodIncrement,powerSpectralDensityOut[x],standardDeviationOut[x]);
	}
	
	/* close up */
	fclose(outputFile);
				
	
	return 0;
}








